//
//  CustomCamera.m
//  iOSDemo
//
//  Created by justinsong on 2021/3/18.
//

#import <Foundation/Foundation.h>
#import "CameraView.h"
#import <AVFoundation/AVFoundation.h>
#import <AVFoundation/AVCaptureDevice.h>
#import <AVFoundation/AVMediaFormat.h>
#import <Accelerate/Accelerate.h>

#define CSizeMake(h, w) [[CSize alloc] initWithHeight: h width: w ]

@interface CSize: NSObject
@property (nonatomic, assign) int width;
@property (nonatomic, assign) int height;
@end

@implementation CSize

// 记得高在前面
- (instancetype)initWithHeight:(int) height width: (int)width{
    self = [super init];
    if (self) {
        _width = width;
        _height = height;
    }
    return self;
}

@end

static NSDictionary *s_presets = @{
    AVCaptureSessionPreset1920x1080: CSizeMake(1920, 1080),
    AVCaptureSessionPreset1280x720: CSizeMake(1280, 720),
    AVCaptureSessionPresetiFrame960x540: CSizeMake(960, 540),
    AVCaptureSessionPreset640x480: CSizeMake(640, 480),
    AVCaptureSessionPreset352x288: CSizeMake(352, 288),
};

@interface CameraView ()<AVCaptureVideoDataOutputSampleBufferDelegate>
{
    CVPixelBufferRef _pixelBuffer;
}

@property (assign, nonatomic) BOOL hasInit;
@property (assign, nonatomic) int channel;

@property (strong, nonatomic) AVCaptureSession *session;
@property (strong, nonatomic) AVCaptureDevice *captureDevice;
@property (strong, nonatomic) AVCaptureDeviceInput *deviceInput;
@property (strong, nonatomic) AVCaptureVideoPreviewLayer *previewLayer;
@property (nonatomic, assign) int width;
@property (nonatomic, assign) int height;
@property (copy, nonatomic) OnGetFrame frameCallBack;

@end


@implementation CameraView

- (void)dealloc {
    
}


#pragma mark - init func


/** 初始化摄像头*/
- (void)init:(int) width height: (int)height {
    
    if (self.hasInit && self.width == width && self.height == height) {
        // 宽高一样的时候, 不用重新设置
        return;
    }
    NSLog(@"CustomCamera, do init camera");
    self.hasInit = YES;
    
    // 创建前置摄像头
    [self initCaptureDevice];
    
    // 创建session，設置分辨率
    [self initSession: width height: height];
    
    // 创建预览，并关联到CamerView上
    [self initPreview];

    self.channel = 4;
}

- (void)uninit{
    self.hasInit = NO;
    self.session = nil;
}
/** 配置摄像头*/
- (void)initCaptureDevice
{
    // 创建前置摄像头
    NSArray *devices = [AVCaptureDevice devicesWithMediaType:AVMediaTypeVideo];
    for (AVCaptureDevice *device in devices)
    {
        if ([device position] == AVCaptureDevicePositionFront)
        {
            self.captureDevice = device;
        }
    }
    
    AVCaptureDevice * device = self.captureDevice ;
    if ([device lockForConfiguration:nil]) {
        // 对焦范围
        if (device.autoFocusRangeRestrictionSupported) {
            device.autoFocusRangeRestriction = AVCaptureAutoFocusRangeRestrictionNear;
        }
        // 光滑对焦
        if (device.smoothAutoFocusSupported) {
            device.smoothAutoFocusEnabled = YES;
        }
        // 智能自动对焦
        if([device isFocusModeSupported:AVCaptureFocusModeContinuousAutoFocus]){
            device.focusMode = AVCaptureFocusModeContinuousAutoFocus;
        }
        // 智能调整曝光
        device.exposureMode = AVCaptureExposureModeContinuousAutoExposure;
        
        [device unlockForConfiguration];
    }
}

- (AVCaptureSessionPreset) selectSizeByHeight:(int) height width: (int)width{
    AVCaptureSessionPreset match  = nil;
    
    for(NSString *p in s_presets){
        CSize *size = [s_presets objectForKey: p];
        if(width == size.width && height == size.height){
            match = p;
            break;
        }
    }
    if(match != nil){
        return match;
    }
    // 得到与传入的宽高比最接近的size
    float reqRatio = ((float) width) / height;
    float curRatio, deltaRatio;
    float deltaRatioMin = NSIntegerMax;
    
    int tmpWidth, tmpHeight;
    for (NSString *p in s_presets) {
        CSize *size = [s_presets objectForKey:p];
        tmpWidth = size.width;
        tmpHeight = size.height;
        curRatio = ((float) tmpWidth) / tmpHeight;
        deltaRatio = abs(reqRatio - curRatio);
        if (deltaRatio < deltaRatioMin) {
            deltaRatioMin = deltaRatio;
            match = p;
        }
    }
    return match;
}

/** 创建session，設置分辨率*/
- (void)initSession:(int) width height: (int)height
{
    // 创建session
    self.session = [[AVCaptureSession alloc] init];
    
    // 分辨率, 从几个之中选择, 不能任意指定
    AVCaptureSessionPreset preset = [self selectSizeByHeight:height width:width];
    self.session.sessionPreset = preset;
    CSize * size = [s_presets objectForKey: preset];
    self.height = size.height;
    self.width = size.width;
    NSLog(@"CustomCamera, select preset: %s, size(width=%d,height=%d)", preset.UTF8String, self.width, self.height);
}

/** 创建预览，并关联到自身*/
- (void)initPreview {
    self.previewLayer = [[AVCaptureVideoPreviewLayer alloc] initWithSession:self.session];
    [self.previewLayer setVideoGravity:AVLayerVideoGravityResizeAspectFill];
    CALayer *rootLayer = [self layer];
    [rootLayer setMasksToBounds:YES];
    CGRect frame = self.frame;
    [self.previewLayer setFrame:frame];
}

/** 设置相机输出*/
- (void)initCameraOutput {
    NSLog(@"CustomCamera, initCameraOutput.");
    AVCaptureVideoDataOutput *videoDataOutput = [AVCaptureVideoDataOutput new];
    
    NSDictionary *rgbOutputSettings = [NSDictionary
                                       dictionaryWithObject:[NSNumber numberWithInt:kCVPixelFormatType_32BGRA]
                                       forKey:(id)kCVPixelBufferPixelFormatTypeKey];
    [videoDataOutput setVideoSettings:rgbOutputSettings];
    [videoDataOutput setAlwaysDiscardsLateVideoFrames:YES];
    dispatch_queue_t videoDataOutputQueue = dispatch_queue_create("VideoDataOutputQueue", DISPATCH_QUEUE_SERIAL);
    [videoDataOutput setSampleBufferDelegate:self queue:videoDataOutputQueue];
    
    if ([self.session canAddOutput:videoDataOutput]){
        [self.session addOutput:videoDataOutput];
    }else{
        NSLog(@"CustomCamera, videoDataOutput add fail!!!");
    }
    [[videoDataOutput connectionWithMediaType:AVMediaTypeVideo] setEnabled:YES];
    
    AVCaptureConnection *connection = [videoDataOutput connectionWithMediaType: AVMediaTypeVideo];
    
    if (connection == nil) {
        [_session commitConfiguration];
        NSLog(@"CustomCamera, connection==nil.");
        return;
    }
    
    // 因为 OpenGL 的纹理 Y 轴和 UIKit 的是相反的，所以这里采集需要上下颠倒
    connection.videoOrientation = AVCaptureVideoOrientationPortrait;
    connection.videoMirrored = YES;
    NSLog(@"CustomCamera, initCameraOutput end.");
}

- (void)addNotification {
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(captureSessionDidFailWithError:) name:AVCaptureSessionRuntimeErrorNotification object:self.session];
}

- (void)removeNotification {
    [[NSNotificationCenter defaultCenter] removeObserver:self];
}


#pragma mark - public func

/** 开始录制, 必须提前 init*/
- (bool)start:(OnGetFrame)frameCallBack {
    
    // 关联session和摄像头，创建AVCaptureDeviceInput会触发请求相机权限
    NSError *error;
    self.deviceInput = [AVCaptureDeviceInput deviceInputWithDevice:self.captureDevice error:&error];
    if ([self.session canAddInput:self.deviceInput]) {
        [self.session addInput:self.deviceInput];
    }
    
    AVAuthorizationStatus status = [AVCaptureDevice authorizationStatusForMediaType:AVMediaTypeVideo];
    if (status == AVAuthorizationStatusRestricted || status == AVAuthorizationStatusDenied)
    {
        // 无权限
        NSLog(@"CustomCamera, start: Error, have no access to open Camera.");
        return false;
    }
    
    // 设置相机输出
    [self initCameraOutput];
    
    self.frameCallBack = frameCallBack;
    [self.session startRunning];
    return true;
}

/** 结束录制*/
- (void)stop {
    self.frameCallBack = nil;
    [self.session stopRunning];
    [self uninit];
    
}

#pragma mark - AVCaptureVideoDataOutputSampleBufferDelegate

- (void)captureOutput:(AVCaptureOutput *)captureOutput didOutputSampleBuffer:(CMSampleBufferRef)sampleBuffer fromConnection:(AVCaptureConnection *)connection {
    @synchronized(self) {
        
        @autoreleasepool {
            
            if (self.frameCallBack) {
                CVPixelBufferRef pixelBuffer = CMSampleBufferGetImageBuffer(sampleBuffer);
                OSType type = CVPixelBufferGetPixelFormatType(pixelBuffer);
                ZegoImageParam *param = [[ZegoImageParam alloc] init];
                if(type == 'BGRA'){
                    param.format = ZegoImageFormatBGRA;
                }else if(type == '420f'){
                    param.format = ZegoImageFormatYUVNV12;
                }
                param.width = self.width;
                param.height = self.height;
                param.rotation = 0;
                param.isMirrored = NO;
                NSTimeInterval start = [[NSDate date] timeIntervalSince1970] * 1000;
                param.timestamp = start;
                
                self.frameCallBack(pixelBuffer, param);
            }
        }
    }
}



- (void)captureSessionDidFailWithError:(NSNotification*)notification {
    NSLog(@"CustomCamera, captureSessionDidFailWithError: %@", notification);
    
}
 

@end
