//
//  MetalTools.m
//  GoAvatar
//
//  Created by horacezhou on 2022/6/27.
//

#import "MetalTools.h"

#define STRINGIZE(x) #x
#define STRINGIZE2(x) STRINGIZE(x)
#define SHADER_STRING(text) @ STRINGIZE2(text)


NSString *const kFlagmentShaderString = SHADER_STRING
(
 
 //#include <metal_stdlib>
 using namespace metal;
 
 struct VertexOut
 {
  float4 position [[position]];
  float2 textureCoordinate;
  float4 color;
};
 
 vertex VertexOut texture_vertex (
                                  constant float4*vertex_array[[buffer(0)]],
                                  constant float2*textureCoord_array[[buffer(1)]],
                                  constant float4*color_array[[buffer(2)]],
                                  unsigned int vid[[vertex_id]]) {
                                    
                                    VertexOut outputVertices;
                                    
                                    outputVertices.position = vertex_array[vid];
                                    outputVertices.textureCoordinate = textureCoord_array[vid];
                                    outputVertices.color = color_array[0];
                                    
                                    return outputVertices;
                                  }
 
 inline float GammaToLinearSpaceExact (float value)
 {
  if (value <= 0.04045F)
    return value / 12.92F;
  else if (value < 1.0F)
    return pow((value + 0.055F)/1.055F, 2.4F);
  else
    return pow(value, 2.2F);
}
 
 fragment float4 texture_fragment(VertexOut fragmentInput [[stage_in]],
                                  texture2d<float> inputTexture [[texture(0)]]) {
  constexpr sampler quadSampler;
  float4 rgb = inputTexture.sample(quadSampler, fragmentInput.textureCoordinate).bgra;
  
  //float3 gamma = fragmentInput.color.rgb;
  //float4 line = float4( GammaToLinearSpaceExact(gamma.r),GammaToLinearSpaceExact(gamma.g),GammaToLinearSpaceExact(gamma.b),1.0); //RGB互换一下
  float4 backColor = fragmentInput.color.bgra;
  return float4(rgb * rgb.a + backColor * (1.0 - rgb.a));
}
 
 
 );

@interface MetalTools()
{
  id <MTLDevice>                  _device;
  id <MTLRenderPipelineState>     _pipelineState;
  MTLRenderPassDescriptor*        _renderPassDes;
  id <MTLCommandQueue>            _commandQueue;
  MTLTextureDescriptor*           _texDescriptor;
}

@end

@implementation MetalTools

+ (MetalTools *)sharedInstance
{
    static MetalTools *tools = nil;
    static dispatch_once_t onceToken;
    dispatch_once(&onceToken, ^{
        tools = [[MetalTools alloc] init];
    });

    return tools;
}

- (id)init{
  
  if (self = [super init]) {
    [self setupPineline];
  }
  
  return self;
}

-(void)dealloc
{
  
}

- (void)setupBgColorWithTexture:(id<MTLTexture>)texture bgColor:(UIColor *)bgColor callback:(OnMetalCallback) metalCallBack{
  
  _texDescriptor.width = texture.width;
  
  _texDescriptor.height = texture.height;
  
  _texDescriptor.pixelFormat = texture.pixelFormat;
  
  id<MTLTexture> newTexture = [_device newTextureWithDescriptor:_texDescriptor];
  
  MTLRegion region = MTLRegionMake2D(0, 0, texture.width, texture.height);
  
  id <MTLCommandBuffer> commandBuffer = [_commandQueue commandBuffer];
  
  id <MTLBlitCommandEncoder> commandEncoder = commandBuffer.blitCommandEncoder;
  
  if (@available(iOS 13.0, *)) {
    [commandEncoder copyFromTexture:texture toTexture:newTexture];
  } else {
    [commandEncoder copyFromTexture:texture sourceSlice:0 sourceLevel:0 sourceOrigin:region.origin sourceSize:region.size toTexture:newTexture destinationSlice:0 destinationLevel:0 destinationOrigin:region.origin];
  }
  
  [commandEncoder endEncoding];
    commandEncoder = nil;
  
  _renderPassDes.colorAttachments[0].texture = newTexture;
  
  id<MTLRenderCommandEncoder> renderEncoder = [commandBuffer renderCommandEncoderWithDescriptor:_renderPassDes];
  
    
  __weak id <MTLTexture> weakTexture = newTexture;
  [commandBuffer addCompletedHandler:^(id<MTLCommandBuffer> _Nonnull buf) {
    if(metalCallBack){
      __strong  id<MTLTexture> strongTexture = weakTexture;
      metalCallBack(strongTexture);

    }
  }];
  
  renderEncoder.label = @"Offscreen Render Pass AvatarExpress";
  
  float vertexArray[] = {
    -1.0, -1.0, 0.0, 1.0,
    1.0, -1.0, 0.0, 1.0,
    -1.0,  1.0, 0.0, 1.0,
    1.0,  1.0, 0.0, 1.0,
  };
  
  id<MTLBuffer> vertexBuffer = [_device newBufferWithBytes:vertexArray length:sizeof(vertexArray) options:MTLResourceCPUCacheModeDefaultCache];
  
  [renderEncoder setVertexBuffer:vertexBuffer offset:0 atIndex:0];
  
  NSInteger a = sizeof(vertexArray) / (sizeof(float)*4);
  
  float textureCoord[] = {
    0,1,
    1,1,
    0,0,
    1,0
  };
  id<MTLBuffer>textureCoordBuffer = [_device newBufferWithBytes:textureCoord length:sizeof(textureCoord) options:MTLResourceCPUCacheModeDefaultCache];
  
  [renderEncoder setVertexBuffer:textureCoordBuffer offset:0 atIndex:1];
  
  //    UIColor *bgColor = self.bgColor;
  CGFloat redCG = 1.0;
  CGFloat greenCG = 1.0;
  CGFloat blueCG = 1.0;
  CGFloat alphaCG = 1.0;
  if (bgColor) {
    [bgColor getRed:&redCG green:&greenCG blue:&blueCG alpha:&alphaCG];
  }
  float red = redCG;
  float green = greenCG;
  float blue = blueCG;
  float alpha = alphaCG;
  
  float colorCoord[] = {
    red, green, blue, alpha,
  };
  
  id<MTLBuffer> colorCoordBuffer = [_device newBufferWithBytes:colorCoord length:sizeof(colorCoord) options:MTLResourceCPUCacheModeDefaultCache];
  
  [renderEncoder setVertexBuffer:colorCoordBuffer offset:0 atIndex:2];
  
  [renderEncoder setFragmentTexture:newTexture atIndex:0];
  [renderEncoder setRenderPipelineState:_pipelineState];
  [renderEncoder drawPrimitives:MTLPrimitiveTypeTriangleStrip vertexStart:0 vertexCount:a];
  [renderEncoder endEncoding];
  [commandBuffer commit];
}

- (void)setupPineline {
  _device = MTLCreateSystemDefaultDevice();
  
  _commandQueue = [_device newCommandQueue];
  
  
  id<MTLLibrary>library ;
  NSError *error;
  MTLCompileOptions* options = [[MTLCompileOptions alloc] init];
  // Runtime checks for the iOS version independently of minimum target iOS.
  if (@available(macOS 10.14, iOS 12.0, tvOS 12.0, *)) {
    [options setLanguageVersion:MTLLanguageVersion2_1];
  } else if (@available(macOS 10.13, iOS 11.0, tvOS 11.0, *)) {
    [options setLanguageVersion:MTLLanguageVersion2_0];
  } else if (@available(macOS 10.12, iOS 10.0, tvOS 10.0, *)) {
    [options setLanguageVersion:MTLLanguageVersion1_2];
  } else if (@available(macOS 10.11, iOS 9.0, tvOS 9.0, *)) {
    [options setLanguageVersion:MTLLanguageVersion1_1];
  }
  NSString* shader = [NSString stringWithFormat:@"#include <metal_stdlib> \n\n%@",kFlagmentShaderString];
  
  library = [_device newLibraryWithSource:shader options:options error:&error];
  
  
  // 初始化pipelineState
  MTLRenderPipelineDescriptor *pinelineDesc = [MTLRenderPipelineDescriptor new];
  pinelineDesc.vertexFunction = [library newFunctionWithName:@"texture_vertex"];
  pinelineDesc.fragmentFunction = [library newFunctionWithName:@"texture_fragment"];
  pinelineDesc.colorAttachments[0].pixelFormat = MTLPixelFormatBGRA8Unorm;
  _pipelineState = [_device newRenderPipelineStateWithDescriptor:pinelineDesc error:&error];
  
  MTLRenderPassDescriptor *renderPassDes = [[MTLRenderPassDescriptor alloc] init];
  renderPassDes.colorAttachments[0].storeAction = MTLStoreActionStore;
  renderPassDes.colorAttachments[0].loadAction = MTLLoadActionClear;
  renderPassDes.colorAttachments[0].clearColor = MTLClearColorMake(0.0, 0.0, 0.0, 0.0);
  _renderPassDes = renderPassDes;
  
  MTLTextureDescriptor *texDescriptor = [MTLTextureDescriptor new];
  texDescriptor.textureType = MTLTextureType2D;
  texDescriptor.usage = MTLTextureUsageRenderTarget | MTLTextureUsageShaderRead;
  _texDescriptor = texDescriptor;
}
@end
