#import "ZegoTextureRenderer.h"

@interface ZegoTextureRenderer ()
@property (nonatomic, strong) CAMetalLayer *layer;
@property (nonatomic, strong) ZegoMetalContext *metalContext;

@property (nonatomic, strong) id<MTLSamplerState> samplerState;
@property (nonatomic, strong) id<MTLRenderPipelineState> textureRenderPipeline;
@property (nonatomic, strong) id<MTLBuffer> textureVertexBuffer;
@end

@implementation ZegoTextureRenderer

- (instancetype)initWithLayer:(CAMetalLayer *)layer andContext: (ZegoMetalContext *)context{
    if ((self = [super init])){
        _layer = layer;
        _metalContext=context;
        self.fillMode = ZegoCaptureVideoFillMode_GravityResize;
        
        [self buildMetal];
        [self buildPipelines];
        [self buildResources];
    }
    return self;
}

- (void)buildMetal{
    _layer.device = _metalContext.device;
    _layer.pixelFormat = MTLPixelFormatBGRA8Unorm;
}

- (void)buildPipelines{
    NSError *error = nil;
    id<MTLLibrary> library = _metalContext.library;
    
    id<MTLFunction> textureVertexFunc = [library newFunctionWithName:@"texture_vertex_main"];
    id<MTLFunction> textureFragmentFunc = [library newFunctionWithName:@"texture_fragment_main"];
    
    MTLRenderPipelineDescriptor *texturePipelineDescriptor = [MTLRenderPipelineDescriptor new];
    texturePipelineDescriptor.colorAttachments[0].pixelFormat = MTLPixelFormatBGRA8Unorm;
    texturePipelineDescriptor.vertexFunction = textureVertexFunc;
    texturePipelineDescriptor.fragmentFunction = textureFragmentFunc;
    
    _textureRenderPipeline = [_metalContext.device newRenderPipelineStateWithDescriptor:texturePipelineDescriptor
                                                                                  error:&error];
    
    if (!_textureRenderPipeline){
        NSLog(@"Error occurred when creating texture render pipeline state: %@", error);
    }
}

- (void)buildResources{
    const float textureVertices[] =
    {
        -1.0, -1.0, 0,
        -1.0,  1.0, 0,
        1.0, -1.0, 0,
        1.0,  1.0, 0
    };
    _textureVertexBuffer = [_metalContext.device newBufferWithBytes:textureVertices
                                                             length:sizeof(textureVertices)
                                                            options:MTLResourceOptionCPUCacheModeDefault];
    
    // create sampler state
    MTLSamplerDescriptor *samplerDesc = [MTLSamplerDescriptor new];
    samplerDesc.sAddressMode = MTLSamplerAddressModeClampToEdge;
    samplerDesc.tAddressMode = MTLSamplerAddressModeClampToEdge;
    samplerDesc.minFilter = MTLSamplerMinMagFilterNearest;
    samplerDesc.magFilter = MTLSamplerMinMagFilterLinear;
    samplerDesc.mipFilter = MTLSamplerMipFilterLinear;
    _samplerState = [_metalContext.device newSamplerStateWithDescriptor:samplerDesc];
}

-(void)setMentalViewport:(id<MTLRenderCommandEncoder>)encoder
            textureWidth:(unsigned int)width
           textureHeight:(unsigned int)heigh{
    // 视频宽高
    CGFloat videoWidth = width;
    CGFloat videoHeight = heigh;
    
    // 视频高宽比
    CGFloat videoScale = videoHeight/videoWidth;
     
    CGSize drawableSize = _layer.drawableSize;
    
    // 画面同宽比
    CGFloat canvaseScale = (CGFloat)(drawableSize.height) / (CGFloat)(drawableSize.width);
    
    // 视频实际显示的宽高，即在view point中使用的宽高
    CGFloat videoDisplayWidth = drawableSize.width;
    CGFloat videoDisplayHeight = drawableSize.height;
    
    if (self.fillMode == ZegoCaptureVideoFillMode_GravityResize) {
        return;
    }
    
    if (self.fillMode == ZegoCaptureVideoFillMode_GravityResizeAspectFill) {
        if (canvaseScale > videoScale) {
            videoDisplayWidth = (videoWidth/videoHeight) * drawableSize.height;
            videoDisplayHeight = drawableSize.height;
            [encoder setViewport:(MTLViewport){(drawableSize.width - videoDisplayWidth)/2, 0, videoDisplayWidth, videoDisplayHeight, -1.0, 1.0 }];
        }else{
            videoDisplayWidth = drawableSize.width;
            videoDisplayHeight = (videoHeight/videoWidth)*drawableSize.width;
            [encoder setViewport:(MTLViewport){0, (drawableSize.height - videoDisplayHeight)/2, videoDisplayWidth, videoDisplayHeight, -1.0, 1.0 }];
        }
        
        return;
    }
    
    if (self.fillMode == ZegoCaptureVideoFillMode_GravityResizeAspect) {
        if (canvaseScale > videoScale) {
            videoDisplayWidth = drawableSize.width;
            videoDisplayHeight = drawableSize.width * (videoHeight/videoWidth);
            [encoder setViewport:(MTLViewport){0.0, (drawableSize.height - videoDisplayHeight)/2, videoDisplayWidth, videoDisplayHeight, -1.0, 1.0 }];
        }else{
            videoDisplayWidth = drawableSize.height * (videoWidth/videoHeight);
            videoDisplayHeight = drawableSize.height;
            [encoder setViewport:(MTLViewport){(drawableSize.width - videoDisplayWidth)/2, 0, videoDisplayWidth, videoDisplayHeight, -1.0, 1.0 }];
        }
        
        return;

    }
}

- (void)draw: (id<MTLTexture>) inTexture{
    id<CAMetalDrawable> drawable = [_layer nextDrawable];
    
    id<MTLTexture> framebufferTexture = drawable.texture;
    if (!drawable){
        return;
    }
    id<MTLCommandBuffer> commandBuffer = [_metalContext.commandQueue commandBuffer];
    commandBuffer.label = @"TextureRendererCommand";
    
    MTLRenderPassDescriptor *passDescriptor = [MTLRenderPassDescriptor renderPassDescriptor];
    passDescriptor.colorAttachments[0].texture = framebufferTexture;
    passDescriptor.colorAttachments[0].clearColor = MTLClearColorMake(1.0, 1.0, 1.0, 1);
    passDescriptor.colorAttachments[0].storeAction = MTLStoreActionStore;
    passDescriptor.colorAttachments[0].loadAction = MTLLoadActionClear;
    
    id<MTLRenderCommandEncoder> renderEncoder = [commandBuffer renderCommandEncoderWithDescriptor:passDescriptor];
    
    [self setMentalViewport:renderEncoder textureWidth:inTexture.width textureHeight:inTexture.height];
 
    [renderEncoder setRenderPipelineState:_textureRenderPipeline];
    [renderEncoder setFragmentTexture:inTexture atIndex:0];
    [renderEncoder setFragmentSamplerState:_samplerState atIndex:0];
    [renderEncoder setVertexBuffer: _textureVertexBuffer offset:0 atIndex:0];
    [renderEncoder drawPrimitives:MTLPrimitiveTypeTriangle vertexStart:0 vertexCount:3];
    [renderEncoder drawPrimitives:MTLPrimitiveTypeTriangle vertexStart:1 vertexCount:3];
    [renderEncoder endEncoding];
    
    [commandBuffer presentDrawable:drawable];
    [commandBuffer commit];
}

@end
