//
//  AudioQueuePlay.m
//  GoAvatar
//
//  Created by lihua.wei on 2023/2/8.
//

#import "AudioQueuePlay.h"

#define MIN_SIZE_PER_FRAME 0x100000
#define QUEUE_BUFFER_SIZE 3      //队列缓冲个数

@interface AudioQueuePlay() {
    
    AudioQueueRef audioQueue;                                 //音频播放队列
    AudioStreamBasicDescription _audioDescription;
    AudioQueueBufferRef audioQueueBuffers[QUEUE_BUFFER_SIZE]; //音频缓存
    BOOL audioQueueBufferUsed[QUEUE_BUFFER_SIZE];             //判断音频缓存是否在使用
    NSLock *sysnLock;
    NSMutableData *tempData;
    OSStatus osState;
}

@end

@implementation AudioQueuePlay

- (instancetype)init
{
    self = [super init];
    if (self) {
        sysnLock = [[NSLock alloc] init];
        
        // 播放PCM使用
        if (_audioDescription.mSampleRate <= 0) {
            // format
            _audioDescription.mSampleRate       = 16000.0; // 采样率
            _audioDescription.mFormatID         = kAudioFormatLinearPCM; // PCM格式
            _audioDescription.mFormatFlags      = kLinearPCMFormatFlagIsSignedInteger; // 整形
            _audioDescription.mFramesPerPacket  = 1; // 每帧只有1个packet
            _audioDescription.mChannelsPerFrame = 1; // 声道数
            _audioDescription.mBytesPerFrame    = 2; // 每帧只有2个byte 声道*位深*Packet数
            _audioDescription.mBytesPerPacket   = 2; // 每个Packet只有2个byte
            _audioDescription.mBitsPerChannel   = 16; // 位深
        }
        
        // 使用player的内部线程播放 新建输出
        osState = AudioQueueNewOutput(&_audioDescription, AudioPlayerAQInputCallback, (__bridge void * _Nullable)(self), nil, 0, 0, &audioQueue);
        
        // 设置音量
        AudioQueueSetParameter(audioQueue, kAudioQueueParam_Volume, 1.0);
        
        
        
        // 初始化需要的缓冲区
        for (int i = 0; i < QUEUE_BUFFER_SIZE; i++) {
            audioQueueBufferUsed[i] = false;

            osState = AudioQueueAllocateBuffer(audioQueue, MIN_SIZE_PER_FRAME, &audioQueueBuffers[i]);

            printf("第 %d 个AudioQueueAllocateBuffer 初始化结果 %d (0表示成功)", i + 1, osState);
        }
        
        osState = AudioQueueStart(audioQueue, NULL);
        if (osState != noErr) {
            printf("AudioQueueStart Error");
        }
    }
    return self;
}

- (void)resetPlay {
    if (audioQueue != nil) {
        AudioQueueReset(audioQueue);
    }
}

// 播放相关
-(void)playWithData:(NSData *)data {
    
    [sysnLock lock];
    
    tempData = [NSMutableData new];
    [tempData appendData:data];
    // 得到数据
    NSUInteger len = tempData.length;
    Byte *bytes = (Byte*)malloc(len);
    [data getBytes:bytes length:len];
    
    int i = 0;
    while (true) {
        if (!audioQueueBufferUsed[i]) {
            audioQueueBufferUsed[i] = true;
            break;
        } else {
            i++;
            if (i >= QUEUE_BUFFER_SIZE) {
                i = 0;
            }
        }
    }
    
    audioQueueBuffers[i] -> mAudioDataByteSize = (UInt32)len;
    // 把bytes的头地址开始的len字节给mAudioData
    memcpy(audioQueueBuffers[i] -> mAudioData, bytes, len);
    
    free(bytes);
    OSStatus status = AudioQueueEnqueueBuffer(audioQueue, audioQueueBuffers[i], 0, NULL);
    osState = AudioQueueStart(audioQueue, NULL);
    
    printf("本次播放数据大小: %lu, %d\n", len, status);
    [sysnLock unlock];
}

// ************************** 回调 **********************************

// 回调回来把buffer状态设为未使用
static void AudioPlayerAQInputCallback(void* inUserData,AudioQueueRef audioQueueRef, AudioQueueBufferRef audioQueueBufferRef) {
    
    AudioQueuePlay *player = (__bridge AudioQueuePlay*)inUserData;
    
    [player resetBufferState:audioQueueRef and:audioQueueBufferRef];
}

- (void)resetBufferState:(AudioQueueRef)audioQueueRef and:(AudioQueueBufferRef)audioQueueBufferRef {
    
    for (int i = 0; i < QUEUE_BUFFER_SIZE; i++) {
        // 将这个buffer设为未使用
        if (audioQueueBufferRef == audioQueueBuffers[i]) {
            audioQueueBufferUsed[i] = false;
        }
    }
}

// ************************** 内存回收 **********************************

- (void)dealloc {
    
    if (audioQueue != nil) {
        AudioQueueStop(audioQueue,true);
    }
    
    audioQueue = nil;
    sysnLock = nil;
}


- (void)stop {
    if (audioQueue != nil) {
        AudioQueueReset(audioQueue);
        AudioQueueStop(audioQueue,true);
    }
}

@end
