//
//  ZegoTextAPI.m
//  GoAvatar
//
//  Created by lihua.wei on 2023/2/7.
//

#import "ZegoTextAPI.h"
#import "AudioQueuePlay.h"
#import "GTMBase64.h"
#import <AFNetworking/AFNetworkReachabilityManager.h>
#import "ZegoTextLicenseCustomer.h"
#import <YYKit/NSString+YYAdd.h>
#import <ZegoAvatar/ZegoTypes.h>
#import <ZegoAvatar/IZegoCharacter.h>

// 按大于50个字的第一个字符分割文本
#define kSubTextLength 50

// 音频播放参数
#define kSampleRate 16000.0
#define kBytesPerFrame 2.0
#define kFrame 1.0/30.0

#define kResetBsInterval 0.03 * NSEC_PER_SEC

@interface ZegoAudioModel : NSObject

@property (nonatomic, assign) NSTimeInterval start;
@property (nonatomic, strong) NSArray *bsList;
@property (nonatomic, assign) NSTimeInterval duration;

@end

@implementation ZegoAudioModel

@end


@interface ZegoTextAPI () {
    id<IZegoCharacter> _character;
}

@property (nonatomic, strong) dispatch_queue_t requestQueue;

// 音频播放器
@property (nonatomic, strong) AudioQueuePlay *audioPlayer;
@property (nonatomic, strong) NSTimer *audioPlayTimer;
@property (nonatomic, strong) dispatch_queue_t audioPlayQueue;
@property (nonatomic, assign) BOOL stopAudioPlay;


// 音频播放model
@property (nonatomic, strong) NSMutableArray *audioModelList;
@property (nonatomic, strong) ZegoTextModel *audioCurrentModel;

// 当前表情index
@property (nonatomic, assign) NSUInteger currentBsIndex;
// 最后一帧表情数据
@property (nonatomic, strong) ZegoExpression *lastExpress;
// 重置表情timer
@property (nonatomic, strong) dispatch_source_t resetBsTimer;
// 重置表情index
@property (nonatomic, assign) NSUInteger resetBsIndex;

@end

@implementation ZegoTextAPI

- (instancetype)initWithCharacter:(id<IZegoCharacter>)character {
    self = [super init];
    if (self) {
        _character = character;
        
        self.requestQueue = dispatch_queue_create("text.requst.queue", NULL);
        
        self.audioPlayQueue = dispatch_queue_create("audio.play.queue", NULL);
        self.audioPlayTimer = [NSTimer timerWithTimeInterval:0.01 target:self selector:@selector(audioPlayTimerTick) userInfo:nil repeats:YES];
        [[NSRunLoop mainRunLoop] addTimer:self.audioPlayTimer forMode:NSRunLoopCommonModes];
        [self.audioPlayTimer setFireDate:[NSDate distantFuture]];
        
    }
    return self;
}

// 设置代理
- (void)setDelegate:(id<ITextExpressionDelegate>)delegate {
    _delegate = delegate;
}

// 请求并播放
- (void)playTextExpression:(NSString *)text {
    [self playTextExpression:text speed:5 voiceType:11002];
}

// 停止播放
- (void)stopTextExpression {
    [self stopPlay];
}

// 请求并播放
- (void)playTextExpression:(NSString *)text speed:(NSInteger)speed voiceType:(NSInteger)voiceType {
    
    if (_character == nil && [self.delegate respondsToSelector:@selector(onError:message:)]) {
        [self.delegate onError:ZegoTextAPIErrorCode_ErrInvalidCharacterNil message:@"character为空"];
        return;
    }
    
    if (text.length == 0 && [self.delegate respondsToSelector:@selector(onError:message:)]) {
        [self.delegate onError:ZegoTextAPIErrorCode_ErrInvalidTextNil message:@"文本为空"];
        return;
    } else if (text.length > 1000 && [self.delegate respondsToSelector:@selector(onError:message:)]) {
        [self.delegate onError:ZegoTextAPIErrorCode_ErrInvalidTextOverlong message:@"文本应小于1000个字符"];
        return;
    }
    
    __block NSString *textStr = text;
    
    // 使用到AFNetworking
    AFNetworkReachabilityManager *manger = [AFNetworkReachabilityManager sharedManager];
    [manger setReachabilityStatusChangeBlock:^(AFNetworkReachabilityStatus status) {
        
        switch (status) {
            case 0: {
                
                if ([self.delegate respondsToSelector:@selector(onError:message:)]) {
                    [self.delegate onError:-1 message:@"无网络连接"];
                }
                textStr = nil;
            }
                
                break;

            default: {
                
                // 重置播放标志位
                self.stopAudioPlay = NO;
                
                // 分割文本
                NSArray *arr = [self splitText:textStr];
                
                dispatch_semaphore_t semaphore = dispatch_semaphore_create(0);
                
                for (NSString *str in arr) {
                    dispatch_async(self.requestQueue, ^{
                        
                        NSDate *datenow = [NSDate date];
                        // 时间戳
                        NSString *timeSp = [NSString stringWithFormat:@"%ld", (long)([datenow timeIntervalSince1970]*1000)];
                        
                        // 签名
                        NSString *signatureNonce = [self randomStringWithLength:16];
                        NSUInteger appID = [ZegoTextLicenseCustomer zegoTextAppID];
                        NSString *appSign = [[ZegoTextLicenseCustomer zegoTextAppSign] substringWithRange:NSMakeRange(0, 32)];
                        NSString *temp = [NSString stringWithFormat:@"%ld%@%@%@",appID,signatureNonce,appSign,timeSp];
                        // 使用到YYKit
                        NSString *signature = [temp md5String];

                        // 请求
                        ZegoTextRequest *request = [[ZegoTextRequest alloc] init];
                        [request requestBsWithText:str
                                             appID:appID
                                           appSign:signature
                                         timestamp:timeSp.longValue
                                         voiceType:voiceType
                                          language:1
                                        sampleRate:16000
                                             codec:@"pcm"
                                             nonce:signatureNonce
                                             speed:speed
                                         completed:^(ZegoTextModel * _Nonnull model) {
                            
                            if ((model == nil || model.code != 0) && [self.delegate respondsToSelector:@selector(onError:message:)]) {
                                [self.delegate onError:model.code message:model.message];
                            } else {
                                [self playAudio:model];
                            }
                            
                            dispatch_semaphore_signal(semaphore);
                        }];
                        dispatch_semaphore_wait(semaphore, DISPATCH_TIME_FOREVER);
                    });
                }
            }
                break;
        }
    }];
    [manger startMonitoring];
}

// 生成len位的随机字符串
- (NSString *)randomStringWithLength:(NSInteger)len {
     NSString *letters = @"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
    NSMutableString *randomString = [NSMutableString stringWithCapacity: len];
    
    for (NSInteger i = 0; i < len; i++) {
        [randomString appendFormat: @"%C", [letters characterAtIndex: arc4random_uniform((uint32_t)[letters length])]];
    }
    return randomString;
}

// 分割文本
- (NSArray *)splitText:(NSString *)text {
    
    NSSet *set = [NSSet setWithObjects:
                  @"，",@"。",@"！",@"？",@"；",
                  @",",@".",@"!",@"?",@";",@" ", nil];
    
    NSMutableArray *mutArr = [NSMutableArray array];
    NSUInteger startIdx = 0;

    NSRange range;
    for (int i = 0; i < text.length; i++) {
        range = [text rangeOfComposedCharacterSequenceAtIndex:i];
        NSString *currentChar = [text substringWithRange:range];
        if ([set containsObject:currentChar] && i - startIdx + 1 >= kSubTextLength) {
            NSString *subStr = [text substringWithRange:NSMakeRange(startIdx, i - startIdx + 1)];
            [mutArr addObject:subStr];
            startIdx = i + 1;
        }
    }
    
    if (text.length > startIdx) {
        NSString *subStr = [text substringWithRange:NSMakeRange(startIdx, text.length - startIdx)];
        [mutArr addObject:subStr];
    }
    
    return [mutArr copy];
}

#pragma mark - 音频播放
- (void)playAudio:(ZegoTextModel *)model {
    if (self.stopAudioPlay) {
        return;
    }
    if (model == nil) {
        return;
    }
    if (model.code != 0) {
        return;
    }
    [self.audioModelList addObject:model];
    [self.audioPlayTimer setFireDate:[NSDate distantPast]];
}

- (void)audioPlayTimerTick {
    if (self.stopAudioPlay) {
        return;
    }
    
    // 当前时间戳
    NSTimeInterval currentTime = [self getNowTimeTimestamp];
    
    // 当前音频model为空
    if (self.audioCurrentModel == nil) {
        // 音频model大于0
        if (self.audioModelList.count > 0) {
            
            // 获取音频model
            NSData *data = nil;
            self.audioCurrentModel = [self getNextAudioModelWithTime:currentTime data:&data];
            
            // 播放音频
            dispatch_async(self.audioPlayQueue, ^{
                [self.audioPlayer playWithData:data];
            });
        } else {
            // 停止播放
            [self stopPlay];
        }
    } else {
        // 当前时间大于（音频开始播放 - 音频时长）
        if (currentTime - self.audioCurrentModel.data.start > self.audioCurrentModel.data.duration) {

            // 获取下一个音频model
            if (self.audioModelList.count > 0) {
                // 获取音频model
                NSData *data = nil;
                self.audioCurrentModel = [self getNextAudioModelWithTime:currentTime data:&data];
                
                // 播放音频
                dispatch_async(self.audioPlayQueue, ^{
                    [self.audioPlayer playWithData:data];
                });
            } else {
                // 停止播放
                [self stopPlay];
            }
        } else {
            // 当前音频开始时间差
            NSTimeInterval time = currentTime - self.audioCurrentModel.data.start;
            // 当前需要播放表情的index
            NSInteger index = time / self.audioCurrentModel.data.frame;
            
            // 同一个表情index, 不设置
            if (index == self.currentBsIndex) {
                return;
            }
            
            // index小于音频表情数
            if (index < self.audioCurrentModel.data.frame_sequence.count) {
                
                // 处理服务端返回的表情数据
                NSArray *bs = [self.audioCurrentModel.data.frame_sequence objectAtIndex:index];
                NSMutableArray *subArr = [NSMutableArray array];
                for (NSNumber *value in bs) {
                    float temp = value.floatValue / 100.0;
                    [subArr addObject:[NSNumber numberWithFloat:temp]];
                }
                NSString *expStr = [self objectToJson:subArr.copy];
                ZegoExpression *expression = [ZegoExpression jsonToExpression:expStr];
                
                // 设置表情
                [_character setExpression:expression];
                
                self.lastExpress = expression;
                self.currentBsIndex = index;
            } else {
               
                // 回到默认表情
                [self resetBs];
            }
        }
    }
}

- (void)resetBs {
    
    if (self.resetBsTimer) {
        NSLog(@"ZegoTextAPI, resetBsTimer: already in the process of sleep animate!");
        return;
    }
    
    // 创建队列，丢到子线程里执行timer
    dispatch_queue_t queue = dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0);
    self.resetBsTimer = dispatch_source_create(DISPATCH_SOURCE_TYPE_TIMER, 0, 0, queue);
    dispatch_source_set_timer(self.resetBsTimer                 /** timer对象*/
                              , DISPATCH_TIME_NOW   /** 延迟多久启动*/
                              , kResetBsInterval /** 间隔时间*/
                              , 0 * NSEC_PER_SEC);
    
    __weak typeof(self) weakSelf = self;
    dispatch_source_set_event_handler(self.resetBsTimer, ^{
        __strong typeof(weakSelf) strongSelf = weakSelf;
        
        // 默认表情
        ZegoExpression *defaultExpression = [[ZegoExpression alloc] init];
        
        // 过度动画
        if (strongSelf.resetBsIndex < 15
             && strongSelf.lastExpress) {
            
            // 用最后一帧真实表情，混合第一帧睡眠表情
            [strongSelf.lastExpress blendExpression:0.9 exp:defaultExpression];
            
            // 应用过度表情
            [strongSelf->_character setExpression:strongSelf.lastExpress];
            
            strongSelf.resetBsIndex++;
        } else {
            // 应用默认表情，防止15帧仍没有回到默认表情
            [strongSelf->_character setExpression:defaultExpression];
            [strongSelf stopResetBsTimer];
        }
    });
    
    // 启动timer
    dispatch_resume(self.resetBsTimer);
}

// 停止重置表情的tiemr
- (void)stopResetBsTimer {
    if (!self.resetBsTimer) {
        return;
    }
    
    dispatch_source_cancel(self.resetBsTimer);
    self.resetBsTimer = nil;
    self.resetBsIndex = 0;
}

- (ZegoTextModel *)getNextAudioModelWithTime:(NSTimeInterval)time data:(NSData **)data {
    ZegoTextModel *model = nil;
    if (self.audioModelList.count > 0) {
        model = self.audioModelList.firstObject;
        model.data.start = time;
        // 解码
        NSData *audioData = [self base64Decoding:model.data.audio];
        // 时长
        NSTimeInterval duration = audioData.length * 1000 / (kSampleRate * kBytesPerFrame);
        model.data.duration = duration;
        
        // 每frame播放一帧表情
        float frame = duration * kFrame / (model.data.frame_sequence.count * kFrame);
        frame = (frame + kFrame * 1000.0) / 2.0;
        model.data.frame = frame;
        
        // 移除第一个model
        [self.audioModelList removeObjectAtIndex:0];
        
        *data = audioData;
    }
    return model;
}

- (void)stopPlay {
    self.stopAudioPlay = YES;
    [self.audioPlayer stop];
    [self.audioPlayTimer setFireDate:[NSDate distantFuture]];
    self.audioCurrentModel = nil;
    self.currentBsIndex = 0;
    [self resetBs];
    [self.audioPlayer resetPlay];
    if ([self.delegate respondsToSelector:@selector(onEnd)]) {
        [self.delegate onEnd];
    }
}

#pragma mark - 获取当前时间戳
- (NSTimeInterval)getNowTimeTimestamp {
    NSDate *date = [NSDate dateWithTimeIntervalSinceNow:0];
    NSTimeInterval time = [date timeIntervalSince1970] * 1000;
    return time;
}

#pragma mark - 解码base64
- (NSData *)base64Decoding:(NSString *)string {
    NSData *data = [GTMBase64 decodeData:[string dataUsingEncoding:NSUTF8StringEncoding]];
    return data;
}

#pragma mark - 转json
- (NSString *)objectToJson:(id)obj {
    if (obj == nil) {
        return nil;
    }
    
    NSError *error = nil;
    NSData *jsonData = [NSJSONSerialization dataWithJSONObject:obj options:NSJSONWritingPrettyPrinted error:&error];
 
    if ([jsonData length] && error == nil) {
        return [[NSString alloc] initWithData:jsonData encoding:NSUTF8StringEncoding];
    } else {
        return nil;
    }
}

#pragma mark - 懒加载
- (AudioQueuePlay *)audioPlayer {
    if (!_audioPlayer) {
        _audioPlayer = [[AudioQueuePlay alloc] init];
    }
    return _audioPlayer;
}

- (NSMutableArray *)audioModelList {
    if (!_audioModelList) {
        _audioModelList = [NSMutableArray array];
    }
    return _audioModelList;
}

@end
